import xbmc
import xbmcaddon
import time
import os
import sys

#########################
# Init
#########################
addon = xbmcaddon.Addon()
addon_name = addon.getAddonInfo('name')
addon_id = addon.getAddonInfo('id')
addon_version = addon.getAddonInfo('version')
addon_icon = addon.getAddonInfo('icon')
addon_path = addon.getAddonInfo('path')
addon_cachedir = addon.getAddonInfo('profile')
addon_settings = addon.getSetting

# Add external modules to python path
base_dir = xbmc.translatePath('special://home')
path_to_modules = (base_dir + "/addons/" + addon_id + "/resources/lib")
sys.path.append(path_to_modules)

# General Imorts
import requests
import htsp
import json
from time import *
import time
import sqlite3
import sys
import os.path
import shutil
from optparse import OptionParser
from optparse import OptionGroup

# Imports for mail transfer (considering UTF8 and German Umlaute)
import smtplib
from cStringIO import StringIO
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.header import Header
from email import Charset
from email.generator import Generator

# Unicode Workaround (Umlaute etc.)
import sys
reload(sys)
sys.setdefaultencoding("utf-8")


#########################
# Configuration
#########################

# TVheadend configuration
tvheadend_username = addon_settings('tvheadend_username')
tvheadend_password = addon_settings('tvheadend_password')
tvheadend_server = (addon_settings('tvheadend_server'),9982)

# General
movielist_file = addon_settings('movie_list_file')
if (addon_settings('notification') == "true"):
  activate_notification = True
else:
  activate_notification = False

# Mail configuration
executeMinutes = (60 * int(addon_settings('executeMinutes')))
mail_username = addon_settings('mail_username')
mail_password = addon_settings('mail_password')
mail_smtp_server = addon_settings('mail_smtp_server')
mail_address = [addon_settings('mail_sender_name'), addon_settings('mail_sender')]
mail_recipient = [addon_settings('mail_recipient_name'), addon_settings('mail_recipient')]
mail_subject = addon_settings('mail_subject')

# SQLite DB Filenames
storage = xbmc.translatePath('special://userdata/') + 'addon_data/' + addon_id
db_filename = storage + "/epgEventsMatched.db"
db_backup_filename = storage + "/epgEventsMatched.backup.db"
db_swap_filename = storage + "/epgEventsMatched.swap.db"


def log(description):
  xbmc.log("[DEBUG] '%s v%s': %s" % (addon_name, addon_version, description.encode('ascii', 'ignore')), xbmc.LOGNOTICE)


def notification(title, message, image=addon_icon, displaytime=6000):
  xbmc.executeJSONRPC('{"jsonrpc": "2.0", "method": "GUI.ShowNotification", "params": {"title": "%s", "message": "%s", "image": "%s", "displaytime": %i}, "id": "%s"}' % \
                        (title, message, image, displaytime, addon.getAddonInfo('id')))


def get_settings():
  # Get settings
  tvheadend_username = addon_settings('tvheadend_username')
  tvheadend_password = addon_settings('tvheadend_password')
  tvheadend_server = addon_settings('tvheadend_server')
  movielist_file = addon_settings('movie_list_file')
  if (addon_settings('notification') == "true"):
    activate_notification = True
  else:
    activate_notification = False

  executeMinutes = (60 * int(addon_settings('executeMinutes')))
  mail_username = addon_settings('mail_username')
  mail_password = addon_settings('mail_password')
  mail_smtp_server = addon_settings('mail_smtp_server')
  mail_sender = addon_settings('mail_sender')
  mail_recipient = addon_settings('mail_recipient')
  mail_subject = addon_settings('mail_subject')


def check_movie_file():
  # check if movie file configuration is ok
  if os.path.isfile(movielist_file):
    return True
  else:
    return False

########################
# TVheadend connection
########################
def connect_tvheadend_server(username, password):
    client = htsp.HTSPClient(tvheadend_server)

    log("[*] Say hello ...")
    resp = client.hello()
    log("[-] Server answers:")
    log("    Servername:    " + resp['servername'])
    log("    Serverversion: " + resp['serverversion'])
    log("    HTSP version:  " + str(resp['htspversion']))
    challenge = resp['challenge']

    log("[*] Authenticating ...")
    resp = client.authenticate(username, password)
    log("[*] Authentication succeeded")
    return client



########################
# Init SQLite DB (matched EPG results)
########################
def init_and_connect_DB(db_filename):
    log("[*] Using pysqlite version " + str(sqlite3.version))
    log("[*] Using SQLite version " + str(sqlite3.sqlite_version))

    connection = None

    try:
        connection = sqlite3.connect(db_filename)
        log("[*] Connected to SQLite database " + db_filename)
        with connection:
            cur = connection.cursor()
            cur.execute("DROP TABLE IF EXISTS Events")
            cur.execute("CREATE TABLE Events(eventid INT, title TEXT, description TEXT, start INT, stop INT, channelid INT)")
        connection.commit()
        log("[*] Created new SQLite database " + db_filename)
        return connection
    except sqlite3.Error:
        log("ERROR with SQL-Connection")
        if connection:
            connection.rollback()
        log("[*] Error %s:" % sqlite3.Error.args[0])
        sys.exit(1)


########################
# Disconnect SQLite DB
########################
def disconnectDB(connection):
    if connection:
        connection.close()
    log("[*] Disconnected from DB " + str(db_filename))




########################
# Search in EPG for matches
########################
def fill_DB_with_movies_from_file(movielist_file, connection):
    # Iterate through the Most Wanted Movie List and search in EPG
    with open(movielist_file) as f:
        for line in f:
           # Strip the newline
           line = line.rstrip('\n')
           # Query for EPG data
           log("[*] Searching for regular expression: " + line)
           args = { "query":line,"full":1 }
           client.send("epgQuery", args)
           resp = client.recv()
           if resp:
              log("[*] Match(es) found!")
              # Iterate through all matches and parse it to the queue
              # {'eventId': 78377, 'contentType': 21, 'description': '...', 'title': 'NDR Comedy Contest', 'nextEventId': 78379, 'channelId': 13, 'stop': 1435964400, 'episodeId': 78378, 'start': 1435960800}
              for event in resp['events']:
                  eventid = event['eventId']
                  title = unicode(event['title'])
                  log(title)
                  channelid = event['channelId']
                  start = event['start']
                  stop = event['stop']
                  descr = None # Init
                  if 'description' in event.keys():
                      descr = unicode(event['description'])
                  else:
                      descr = ""
                  with connection:
                      cur = connection.cursor()
                      cur.execute("INSERT INTO Events (eventid,title,description,start,stop,channelid) values (?, ?, ?, ?, ?, ?)",(eventid,title,descr,start,stop,channelid))
                  connection.commit()
           # No regexp match
           else:
                log("[*] No match")

########################
# Get channel information - not available yet (Version 14)
########################
def get_channel_name(channelid):
    connectionString = "http://" + tvheadend_server[0] + ":9981/channels"
    resp = requests.post(connectionString, data={"op":"list"}, auth=(tvheadend_username, tvheadend_password))
    channels = resp.json()
    channels =  channels['entries']
    for channel in channels:
        if channel['chid'] == channelid:
            return channel['name']

    return "Not found"


########################
# Generate HTML output for Mail
########################
def query_and_generate_HTML_from_DB(connection):
    # HTML result for the email
    html_result = ""
    html_result += "<table border =1>\n"
    html_result += "<th>Titel</th><th>Beschreibung<th>Sendezeiten und Kanal</th>\n"

    cur = connection.cursor()
    # Title and description as a unique identifier for one movie
    cur.execute("SELECT DISTINCT title,description FROM Events GROUP BY title,description")
    rows_title_descr = cur.fetchall()

    for row_title_descr in rows_title_descr:
        title = row_title_descr[0]
        descr = row_title_descr[1]

        log("[---------------")
        log("[-] Title:       " + title)
        log("[-] Description: " + descr)

        # HTML: Beginning of the movie section
        html_result += "  <tr>\n"
        html_result += "    <td>" + title + "</td>\n"
        html_result += "    <td>" + descr + "</td>\n"

        # Fetch all attributes for one movie (multiple schedules, channels)
        cur.execute("SELECT DISTINCT start,stop,channelid FROM Events WHERE title=? and description=?",(row_title_descr[0], row_title_descr[1]))
        rows_start_stop_channel = cur.fetchall()

        # HTML: Beginning of the schedules (considering multiple schedules for one movie)
        html_result += "    <td><ul>"
        for row_start_stop_channel in rows_start_stop_channel:
            schedule_start = str(row_start_stop_channel[0])
            schedule_stop  = str(row_start_stop_channel[1])

            schedule_start = time.strftime("%d.%m.%Y | %H:%M",time.localtime(row_start_stop_channel[0]))
            schedule_stop  = time.strftime(" - %H:%M",time.localtime(row_start_stop_channel[1]))

            channel  = row_start_stop_channel[2]

            log("[-] Schedule: " + schedule_start + " to " + schedule_stop + " on channel " + get_channel_name(channel))

            # Only break line when there are multiple schedules and the current schedule is not the last
            if (len(rows_start_stop_channel) > 1) and (row_start_stop_channel != rows_start_stop_channel[len(rows_start_stop_channel)-1]):
                html_result += "<li>" + schedule_start + schedule_stop + " auf Kanal " + get_channel_name(channel) + "</li>\n"
            else:
                html_result += "<li>" + schedule_start + schedule_stop + " auf Kanal " + get_channel_name(channel) + "</li>"
                # HTML: Ending of the schedules section
                html_result += "</ul></td>\n"

        # HTML: Ending of the movie section
        html_result += "  </tr>\n"

    html_result += "</table>\n"
    return html_result


########################
# Check for already transferred movies (prevent sending it twice)
########################
def checkUpdates(con_db, db_backup_filename):

    # 1) Delete already mailed movies from the current search DB
    #    Check if backup DB already exists (history of already mailed movies)
    if os.path.isfile(db_backup_filename):
        try:
            con_db_backup = sqlite3.connect(db_backup_filename)
            log("[*] Connected to SQLite database " + db_backup_filename)

        except sqlite3.Error:
            if con_db_backup:
                con_db_backup.rollback()
            log("[*] Error %s:" % sqlite3.Error.args[0])
            sys.exit(1)

        cur_db = con_db.cursor()
        cur_db_backup = con_db_backup.cursor()

        # Title and description as a unique identifier for one movie
        cur_db.execute("SELECT DISTINCT title,description FROM Events GROUP BY title,description")
        rows_db_title_descr = cur_db.fetchall()
        cur_db_backup.execute("SELECT DISTINCT title,description FROM Events GROUP BY title,description")
        rows_db_backup_title_descr = cur_db_backup.fetchall()

        for row_db_title_descr in rows_db_title_descr:
            title_db = row_db_title_descr[0]
            descr_db = row_db_title_descr[1]

            # Check for each movie (identified as: title, description) if it exists in the backup DB (history)
            for row_db_backup_title_descr in rows_db_backup_title_descr:
                # If title and description are the same (=> duplicate)
                if ((title_db in row_db_backup_title_descr) and (descr_db in row_db_backup_title_descr)):
                    title_db_backup = row_db_backup_title_descr[0]
                    descr_db_backup = row_db_backup_title_descr[1]
                    # Delete already known (and sent) movies from current search (db)
                    cur_db.execute("DELETE FROM Events WHERE title=? and description=?", (title_db, descr_db))
                    log("[*] Record identified as duplicate and removed from result set: " + title_db + " -- " + descr_db)

        # 2) Add new search entries to the backup DB (history)
        cur_db.execute("SELECT DISTINCT title,description FROM Events GROUP BY title,description")
        rows_db_title_descr = cur_db.fetchall()
        cur_db_backup.execute("SELECT DISTINCT title,description FROM Events GROUP BY title,description")
        rows_db_backup_title_descr = cur_db_backup.fetchall()

        for row_db_title_descr in rows_db_title_descr:
            title_db = row_db_title_descr[0]
            descr_db = row_db_title_descr[1]

            # Check for each movie (identified as: title, description) if it exists in the backup DB (history)
            for row_db_backup_title_descr in rows_db_backup_title_descr:
                # If title and description are NOT the same (=> new movie) it is added to the backup DB (history)
                if (title_db not in row_db_backup_title_descr) or (descr_db not in row_db_backup_title_descr):
                    title_db_backup = row_db_backup_title_descr[0]
                    descr_db_backup = row_db_backup_title_descr[1]
                    cur_db_backup.execute("INSERT INTO Events (eventid,title,description,start,stop,channelid) values (?, ?, ?, ?, ?, ?)",("",title_db,descr_db,0,0,0))
                    con_db_backup.commit()
                    log("[*] Record added to the backup db (history): " + title_db + " -- " + descr_db)
        # 3) Diconnect from Backup DB
        disconnectDB(con_db_backup)    
    else:
        log("[*] No backup file yet exists ... creating one")
        shutil.copyfile(db_filename, db_backup_filename)
    # Check (for entries) if new movies were added
    cur_db = con_db.cursor()
    cur_db.execute("SELECT DISTINCT title,description FROM Events GROUP BY title,description")
    # No new movies found
    if (len(cur_db.fetchall()) < 1):
        return False
    # New movies found
    else:
        return True


########################
# Send Email
########################
def sendMail(htmlmsg, mail_address, mail_recipient, mail_subject):

    html = ""
    html += """
        <html>
        <head>
        <style>
        table, td, th {
            border: 1px solid black;
        }

        table {
            border-collapse: collapse;
        }

        th {
            height: 50px;
        }

        table, td, th {
            border: 1px solid green;
        }

        td, th {
            padding: 0.5rem;
        }

        th {
            background-color: green;
            color: white;
        }
        </style>
        </head>
        <body>
        """

    html += htmlmsg
    htmlpart = MIMEText(html, 'html', 'UTF-8')

    msg = MIMEMultipart('alternative')
    msg['Subject'] = "%s" % Header(mail_subject, 'utf-8')
    # Only descriptive part of recipient and sender shall be encoded, not the email address
    msg['From'] = "\"%s\" <%s>" % (Header(mail_address[0], 'utf-8'), mail_address[1])
    msg['To'] = "\"%s\" <%s>" % (Header(mail_recipient[0], 'utf-8'), mail_recipient[1])

    # Attach both parts
    htmlpart = MIMEText(html, 'html', 'UTF-8')
    msg.attach(htmlpart)

    # Create a generator and flatten message object to 'file~@~Y
    str_io = StringIO()
    g = Generator(str_io, False)
    g.flatten(msg)
    # str_io.getvalue() contains ready to sent message

    try:
        smtp = smtplib.SMTP(mail_smtp_server)
        smtp.login(mail_username,mail_password)
        smtp.sendmail(mail_address[1], mail_recipient[1], str_io.getvalue())
        log("Successfully sent email")
    except smtplib.SMTPException:
        log("Error: unable to send email")


########################
# MAIN
########################

if __name__ == "__main__":
  while True:

        # Check movie file
        if not check_movie_file():
          msg = "Movie file not found: %s" % (movielist_file)
          log(msg)
          notification(title=addon_name, message=msg, image=addon_icon, displaytime=6000)

        # In case anything changed (update addon settings)
        get_settings()

        # Connect to TVheadend Server
        client = connect_tvheadend_server(tvheadend_username, tvheadend_password)
       
        # Connect and init new DB
        con = init_and_connect_DB(db_filename)

        # Notification
        msg = "Fetching movies from EPG ..."
        log(msg)
        log(str(activate_notification))
        if activate_notification:
          notification(title=addon_name, message=msg, image=addon_icon, displaytime=6000)

        # Search in EPG for movies and write to database (only in case of matches!)
        fill_DB_with_movies_from_file(movielist_file, con)

        # Check if there were any matches or updates
        # Open old state of DB and delete (old) already transferred matches
        if (checkUpdates(con, db_backup_filename)):

            # Search for matching titles
            htmlmsg = query_and_generate_HTML_from_DB(con)

            # Send Mail
            sendMail(htmlmsg, mail_address, mail_recipient, mail_subject)
        else:
            log("[*] No new movies found ... not sending an email")

        # Disconnect
        disconnectDB(con)
        log("Waiting " + str(executeMinutes) + " seconds ...")
        time.sleep(int(executeMinutes))
